from meitorch.mei import MEI
import meitorch.objective as objective
import meitorch.result as result
import meitorch.tools as tools
from meitorch.linearmei import LinearMEI

__all__ = ['MEI', 'objective', 'LinearMEI', 'result', 'tools']


import warnings
from tqdm import tqdm

import numpy as np
import torch
from scipy.ndimage import label
from numpy.linalg import inv, cholesky


def gauss2d(vx, vy, mu, cov):
    """
    Computes the 2D Gaussian distribution with the given parameters.

    :param vx: x coordinate
    :param vy: y coordinate
    :param mu: mean
    :param cov: covariance
    :return: The Gaussian distribution
    """
    input_shape = vx.shape
    mu_x, mu_y = mu
    v = np.stack([vx.ravel() - mu_x, vy.ravel() - mu_y])
    cinv = inv(cholesky(cov))
    y = cinv @ v
    g = np.exp(-0.5 * (y * y).sum(axis=0))
    return g.reshape(input_shape)


def fit_gauss_envelope(img):
    """
    Given an image, finds a Gaussian fit to the image by treating the square of mean shifted image as the distribution.

    :param img: The image
    :return: The Gaussian distribution
    """
    # scale down the image to 2 dimensions
    while len(img.shape) > 2:
        img = img.mean(axis=0)
    vx, vy = np.meshgrid(np.arange(img.shape[1]), np.arange(img.shape[0]))
    rect = (img - img.mean()) ** 2
    pdf = rect / rect.sum()
    mu_x = (vx * pdf).sum()
    mu_y = (vy * pdf).sum()

    cov_xy = (vx * vy * pdf).sum() - mu_x * mu_y
    cov_xx = (vx ** 2 * pdf).sum() - mu_x ** 2
    cov_yy = (vy ** 2 * pdf).sum() - mu_y ** 2

    cov = np.array([[cov_xx, cov_xy], [cov_xy, cov_yy]])

    g = gauss2d(vx, vy, (mu_x, mu_y), cov)
    mu = (mu_x, mu_y)
    return mu, cov, np.sqrt(g.reshape(img.shape))


def roll(tensor, shift, axis):
    """
    Rolls the tensor along the given axis by the given shift

    :param tensor: The tensor to roll
    :param shift: The shift to apply
    :param axis: The axis on which to shift
    :return: The shifted tensor
    """
    if shift == 0:
        return tensor

    if axis < 0:
        axis += tensor.dim()

    dim_size = tensor.size(axis)
    after_start = dim_size - shift
    if shift < 0:
        after_start = -shift
        shift = dim_size - abs(shift)

    before = tensor.narrow(axis, 0, dim_size - shift)
    after = tensor.narrow(axis, after_start, shift)
    return torch.cat([after, before], axis)


def remove_small_area(mask, size_threshold=50):
    """
    Removes contiguous areas in a thresholded image that is smaller in the number of pixels than size_threshold.

    :param mask: The thresholded image
    :param size_threshold: The threshold
    :return: The thresholded image with small areas removed
    """
    mask_mod = mask.copy()
    label_im, nb_labels = label(mask_mod)
    for v in range(0, nb_labels+1):
        area = label_im == v
        s = np.sum(area)
        if s < size_threshold:
            mask_mod[area] = 0
    return mask_mod





def adjust_img_stats(img, mu, sigma, img_min=0, img_max=255, mask=None, max_gain=6000, min_gain=0.001,
                     base_ratio=1.05, verbose=False):
    """
    Adjusts the image statistics to the given mean and standard deviation.

    :param img: The image to adjust
    :param mu: The target mean
    :param sigma: The target standard deviation
    :param img_min: The minimum pixel intensity allowed
    :param img_max: The maximum pixel intensity allowed
    :param mask: The mask to use
    :param max_gain: The maximum gain to use
    :param min_gain: The minimum gain to use
    :param base_ratio: The base ratio to use
    :param verbose: The verbosity
    :return: The adjusted image
    """
    if mask is None:
        mask = np.ones_like(img)
    mimg = img * mask

    delta = img - mimg.sum() / mask.sum()

    def get_image(delta, offset=0):
        return np.clip((delta * mask) + mu + offset, img_min, img_max)

    sigma_i = get_image(delta).std()
    if sigma_i < 1e-8:
        warnings.warn('Zero standard deviation detected.')
        img = np.clip((delta * mask) + mu, img_min, img_max)  # flat image
        unmasked_img = np.clip(delta + mu, img_min, img_max)
    else:
        gain = sigma / sigma_i

        dir_sign = np.sign(np.log(gain))
        ratio = base_ratio ** dir_sign

        max_gain = min(max(img_max - mu, mu - img_min) / np.min(np.abs(delta)[mask > 0]), max_gain)
        min_gain = min_gain

        v = np.linspace(-30, 30, 1000)[:, None, None]

        imgs = []
        unmasked_imgs = []
        conts = []
        while True:
            if verbose:
                print('Trying gain', gain)
            adj_images = get_image(delta * gain, v)
            pos = np.argmin(np.abs(adj_images.mean(axis=(1, 2)) - mu))
            img = adj_images[pos]
            unmasked_img = np.clip(delta * gain + mu + v[pos], img_min, img_max)

            cont = img.std()
            if verbose:
                print('Got mean and contrast', img.mean(), cont)
            imgs.append(img)
            conts.append(cont)
            unmasked_imgs.append(unmasked_img)
            if (cont - sigma) * dir_sign > 0:
                break
            gain = gain * ratio
            if gain > max_gain or gain < min_gain:
                break

        pos = np.argmin(np.abs(np.array(conts) - sigma))
        img = imgs[pos]
        unmasked_img = unmasked_imgs[pos]
    if verbose:
        print('Selected version with mu={} and std={}'.format(img.mean(), img.std()))
    return img, unmasked_img


def adjust_contrast(img, contrast=-1, mu=-1, img_min=0, img_max=255, force=True, verbose=False, steps=5000):
    """
    Performs contrast adjustment of the image, being mindful of the image value bounds (e.g. [0, 255]). Given the bounds
    the normal shift and scale will not guarantee that the resulting image still has the desired mean luminance
    and contrast.

    :param img: The image to adjust the contrast of
    :param contrast: The desired contrast (RMS)
    :param mu: The desired mean luminance
    :param img_min: The minimum pixel intensity allowed
    :param img_max: The maximum pixel intensity allowed
    :param force: if True, iterative approach is taken to produce an image with the desired stats. This will likely cause
                some pixels to saturate in the upper and lower bounds. If False, then image is scaled simply based on ratio of
                current and desired contrast, and then clipped. This likely results in an image that is higher in contrast
                than the original but not quite at the desired contrast and with some pixel information lost due to clipping.
    :param verbose: If True, prints out progress during iterative adjustment
    :param steps: If force=True, this sets the number of iterative steps to be used in adjusting the image. The larger the
                value, the closer the final image would approach the desired contrast.

    :return:
        adjusted_image - a new image adjusted from the original such that the desired mean/contrast is achieved to the
            best of the configuration.
        clipped - Whether any clipping took place. If True, it indicates that some clipping of pixel intensities occured
            and thus some pixel information was lost.
        actual_contrast - the final contrast of the image reached
    """
    current_contrast = img.std()

    if contrast < 0:
        gain = 1   # no adjustment of contrast
    else:
        gain = contrast / current_contrast

    delta = img - img.mean()
    if mu is None or mu < 0: # no adjustment of mean
        mu = img.mean()

    min_pdist = delta[delta > 0].min()
    min_ndist = (-delta[delta < 0]).min()

    # point beyond which scaling would completely saturate out the image (e.g. all pixels would be completely
    # black or white)
    max_lim_gain = max((img_max - mu) / min_pdist, (mu - img_min) / min_ndist)

    vmax = delta.max()
    vmin = delta.min()

    # maximum gain that could be used without losing image information
    max_gain = min((img_max - mu) / vmax, (img_min-mu) / vmin)

    # if True, we already know that the desired contrast cannot be achieved without losing some pixel information
    # into the saturation regime
    clipped = gain > max_gain

    v = np.linspace(0, (img_max-img_min), steps) # candidates for mean adjustment
    if clipped and force:
        if verbose:
            print('Adjusting...')
        cont = []
        imgs = []
        gains = np.logspace(np.log10(gain), np.log10(max_lim_gain), steps)
        # for each gain, perform offset adjustment such that the mean is equal to the set value
        for g in gains:
            img = delta * g + mu
            img = np.clip(img, img_min, img_max)
            offset = img.mean() - mu # shift in clipped image mean caused by the clipping
            if offset < 0: # pixel values needs to be raised
                offset = -offset
                mask = (img_max-img < v[:, None, None])
                nlow = mask.sum(axis=(1, 2)) # pixels that are closer to the bound than v
                nhigh = img.size - nlow
                # calculate the actual shift in mean that can be achieved by shifting all pixels by v
                # then clipping
                va = ((mask * (img_max-img)).sum(axis=(1, 2)) + v * nhigh) / (nlow + nhigh)

                # find the best candidate offset that achieves closest to the desired shift in the mean
                pos = np.argmin(np.abs(va - offset))
                actual_offset = -v[pos]
            else:
                mask = (img-img_min < v[:, None, None])
                nlow = mask.sum(axis=(1, 2))
                nhigh = img.size - nlow
                # actual shift in mean that can be achieved by shifting all pixels by v
                va = ((mask * (img-img_min)).sum(axis=(1, 2)) + v * nhigh) / (nlow + nhigh)
                pos = np.argmin(np.abs(va - offset))
                actual_offset = v[pos]


            img = img - actual_offset
            img = np.clip(img, img_min, img_max)
            # actual contrast achieved with this scale and adjustment
            c = img.std()
            cont.append(c)
            imgs.append(img)
            if c > contrast:
                break
        loc = np.argmin(np.abs(np.array(cont) - contrast))
        adj_img = imgs[loc]
    else:
        adj_img = delta * gain + mu
        adj_img = np.clip(adj_img, img_min, img_max)
    actual_contrast = adj_img.std()
    return adj_img, clipped, actual_contrast


def adjust_contrast_with_mask(img, img_mask=None, contrast=-1, mu=-1, img_min=0, img_max=255, force=True, verbose=False,
                              mu_steps=500, gain_steps=500):
    """
    A version of the contrast adjustment that is mindful of the mask

    Performs contrast adjustment of the image, being mindful of the image value bounds (e.g. [0, 255]). Given the bounds
    the normal shift and scale will not guarantee that the resulting image still has the desired mean luminance
    and contrast.

    :param img: The image to adjust the contrast of
    :param img_mask: The mask to use for the adjustment. If None, then the entire image is used.
    :param contrast: The desired contrast (RMS)
    :param mu: The desired mean luminance
    :param img_min: The minimum pixel intensity allowed
    :param img_max: The maximum pixel intensity allowed
    :param force: if True, iterative approach is taken to produce an image with the desired stats. This will likely cause
                some pixels to saturate in the upper and lower bounds. If False, then image is scaled simply based on ratio of
                current and desired contrast, and then clipped. This likely results in an image that is higher in contrast
                than the original but not quite at the desired contrast and with some pixel information lost due to clipping.
    :param verbose: If True, prints out progress during iterative adjustment
    :param gain_steps: If force=True, this sets the number of iterative steps to be used in adjusting the image. The larger the
                value, the closer the final image would approach the desired contrast.

    :return:
        adjusted_image - a new image adjusted from the original such that the desired mean/contrast is achieved to the
            best of the configuration.
        clipped - Whether any clipping took place. If True, it indicates that some clipping of pixel intensities occured
            and thus some pixel information was lost.
        actual_contrast - the final contrast of the image reached
    """
    if img_mask is None:
        img_mask = np.ones_like(img)

    def get_mu(x):
        return np.sum(img_mask * x) / np.sum(img_mask)

    def get_sigma(x):
        h, w = x.shape[-2:]
        avg = get_mu(x)
        return np.sqrt(np.sum(img_mask ** 2 * (x - avg) ** 2) / (h * w))

    adj_img = img * img_mask + mu * (1 - img_mask) #adj_img volt
    adj_img = np.clip(adj_img, img_min, img_max)
    mimg = img_mask * img
    test_img = np.clip(mimg - mimg.mean() + mu, img_min, img_max)
    current_contrast = test_img.std()
    if verbose:
        print('Initial contrast:', current_contrast)

    if contrast < 0:
        gain = 1  # no adjustment of contrast
    else:
        gain = contrast / current_contrast

    delta = (img - get_mu(img))  # * bin_mask # only consider deltas in mask region
    if mu is None or mu < 0:  # no adjustment of mean
        mu = get_mu(img)

    min_pdist = delta[delta > 0].min()
    min_ndist = (-delta[delta < 0]).min()

    # point beyond which scaling would completely saturate out the image (e.g. all pixels would be completely black or
    # white)
    max_lim_gain = min(max((img_max - mu) / min_pdist, (mu - img_min) / min_ndist), 100)

    vmax = (delta * img_mask).max()
    vmin = (delta * img_mask).min()

    # maximum gain that could be used without losing image information
    max_gain = min((img_max - mu) / vmax, (img_min - mu) / vmin)


    # if True, we already know that the desired contrast cannot be achieved without losing some pixel information
    # into the saturation regime
    clipped = gain > max_gain
    print('gains', gain , max_gain)

    v = np.linspace(0, (img_max - img_min), mu_steps)  # candidates for mean adjustment

    if clipped and force:
        if verbose:
            print('Adjusting...')
        cont = []
        imgs = []
        gains = np.logspace(np.log10(gain), np.log10(max_lim_gain), gain_steps)
        # for each gain, perform offset adjustment such that the mean is equal to the set value
        for g in tqdm(gains, disable=(not verbose)):
            print('')
            img = delta * g + mu
            img = np.clip(img, img_min, img_max)

            offset = mu - get_mu(img)  # shift in clipped image mean caused by the clipping
            if offset > 0:
                sign = 1
                edge = img_max
            else:
                sign = -1
                edge = img_min

            offset = sign * offset
            mask = (sign * (edge - img) < v[:, None, None])

            nlow = (mask * img_mask).sum(axis=(1, 2))  # effective number of pixels that are closer to the bound than v
            nhigh = img_mask.sum() - nlow

            # calculate the actual shift in mean that can be achieved by shifting all pixels by v
            # then clipping
            va = ((mask * img_mask * sign * (edge - img)).sum(axis=(1, 2)) + (
                    v[:, None, None] * img_mask * (1 - mask)).sum(axis=(1, 2))) / (nlow + nhigh)

            # find the best candidate offset that achieves closest to the desired shift in the mean
            pos = np.argmin(np.abs(va - offset))
            actual_offset = sign * v[pos]

            img = img + actual_offset
            img = np.clip(img, img_min, img_max)
            # actual contrast achieved with this scale and adjustment
            c = get_sigma(img)
            print('contrast now', c)
            cont.append(c)
            imgs.append(img)
            if c > contrast:
                break
        loc = np.argmin(np.abs(np.array(cont) - contrast))
        adj_img = imgs[loc]
    else:
        adj_img = delta * gain + mu

    adj_img = adj_img * img_mask + mu * (1 - img_mask)
    adj_img = np.clip(adj_img, img_min, img_max)
    actual_contrast = adj_img.std()
    return adj_img, clipped, actual_contrast


import numpy as np
from numpy.linalg import svd
from skimage.restoration import denoise_nl_means, estimate_sigma

import kornia
from kornia import filters
import torch
from torch import nn
from abc import ABC, abstractmethod
from .schedules import ConstantSchedule


class Denoiser(nn.Module, ABC):
    @abstractmethod
    def forward(self, noisy_image, step_i):
        pass

    @staticmethod
    def init_p(parameter, iter_n):
        if callable(parameter):
            return parameter(iter_n)
        else:
            return ConstantSchedule(parameter)(iter_n)

    @staticmethod
    def get_denoiser(blur_type, n_iters, **params):
        if blur_type == 'bilateral':
            denoiser = FilterBlur(blur_type=blur_type, n_iters=n_iters, **params)
        elif blur_type == 'gaussian':
            denoiser = FilterBlur(blur_type=blur_type, n_iters=n_iters, **params)
        elif blur_type == 'tv':
            assert 'regularization_scaler' in params.keys(), "regularization_scaler must be specified for tv denoiser"
            assert 'num_iters' in params.keys(), "num_iters must be specified for tv denoiser"
            assert 'lr' in params.keys(), "lr must be specified for tv denoiser"
            denoiser = TVDenoise(n_iters=n_iters, **params)
        elif blur_type == 'wnnm':
            assert 'patch_size' in params.keys(), "patch_size must be specified for wnnm denoiser"
            assert 'delta' in params.keys(), "delta must be specified for wnnm denoiser"
            assert 'c' in params.keys(), "c must be specified for wnnm denoiser"
            assert 'K' in params.keys(), "K must be specified for wnnm denoiser"
            assert 'sigma_n' in params.keys(), "sigma_n must be specified for wnnm denoiser"
            assert 'N_threshold' in params.keys(), "N_threshold must be specified for wnnm denoiser"
            assert 'N_iter' in params.keys(), "N_iter must be specified for wnnm denoiser"
            denoiser = WNNMDenoiser(n_iters=n_iters, **params)
        elif blur_type == 'bm3d':
            assert 'sigma_psd' in params.keys(), "sigma_psd must be specified for bm3d denoiser"
            denoiser = BM3DDenoise(n_iters=n_iters, **params)
        elif blur_type == 'nlm':
            assert 'patch_size' in params.keys(), "patch_size must be specified for nlm denoiser"
            assert 'patch_distance' in params.keys(), "patch_distance must be specified for nlm denoiser"
            assert 'h' in params.keys(), "h must be specified for nlm denoiser"
            assert 'fast_mode' in params.keys(), "fast_mode must be specified for nlm denoiser"
            denoiser = NonLocalMeansDenoise(n_iters=n_iters, **params)
        else:
            raise ValueError("Invalid denoiser type "
                             "(must be 'gaussian', 'bilateral', 'tv', 'filter', 'wnnm', 'bm3d')")
        return denoiser


class TVDenoise(Denoiser):
    def __init__(self, n_iters, regularization_scaler=0.0001, num_iters=500, lr=0.1):
        super(TVDenoise, self).__init__()
        self.l2_term = torch.nn.MSELoss(reduction='mean')
        self.regularization_term = kornia.losses.TotalVariation()

        self.regularization_scaler = self.init_p(regularization_scaler, n_iters)
        self.num_iters = self.init_p(num_iters, n_iters)
        self.lr = lr

    def loss(self, clean_image, noisy_image, step_i):
        return self.l2_term(clean_image, noisy_image) + \
            self.regularization_scaler(step_i) * self.regularization_term(clean_image)

    def forward(self, noisy_image, step_i):
        clean_image = torch.nn.Parameter(data=noisy_image.clone(), requires_grad=True)
        optimizer = torch.optim.Adam([clean_image], lr=self.lr)
        for i in range(self.num_iters(step_i)):
            optimizer.zero_grad()
            loss = self.loss(clean_image, noisy_image, step_i)
            loss.backward()
            optimizer.step()
        return clean_image


class FilterBlur(Denoiser):
    def __init__(self, n_iters, blur_type, **filter_params):
        super(FilterBlur, self).__init__()

        self.blur_type = blur_type

        if blur_type == 'bilateral':
            assert 'kernel_size' in filter_params, "kernel_size must be specified for bilateral filter"
            assert 'sigma_color' in filter_params, "sigma_color must be specified for bilateral filter"
            assert 'sigma_spatial' in filter_params, "sigma_spatial must be specified for bilateral filter"

            def params(step_i):
                return ((torch.tensor(filter_params['kernel_size']), torch.tensor(filter_params['kernel_size'])),
                        self.init_p(filter_params['sigma_color'], n_iters)(step_i),
                        (torch.tensor(self.init_p(filter_params['sigma_spatial'], n_iters)(step_i)),
                         torch.tensor(self.init_p(filter_params['sigma_spatial'], n_iters)(step_i))))

            filter_t = filters.BilateralBlur
        elif blur_type == 'gaussian':
            assert 'sigma' in filter_params, "sigma must be specified for gaussian filter"
            assert 'kernel_size' in filter_params, "kernel size must be specified for gaussian filter"

            def params(step_i):
                return ((torch.tensor(filter_params['kernel_size']), torch.tensor(filter_params['kernel_size'])),
                        (torch.tensor(self.init_p(filter_params['sigma'], n_iters)(step_i)),
                         self.init_p(filter_params['sigma'], n_iters)(step_i)))

            filter_t = filters.GaussianBlur2d
        else:
            raise ValueError("Invalid filter type (must be 'bilateral' or 'gaussian')")

        self.filter = filter_t
        self.params = params

    def forward(self, x, step_i):
        params = self.params(step_i)
        return self.filter(*params)(x)


class NonLocalMeansDenoise(Denoiser):
    def __init__(self, n_iters, patch_size=5, patch_distance=6, h=0.8, fast_mode=True):
        super(NonLocalMeansDenoise, self).__init__()

        self.patch_size = self.init_p(patch_size, n_iters)
        self.patch_distance = self.init_p(patch_distance, n_iters)
        self.h = self.init_p(h, n_iters)
        self.fast_mode = fast_mode

    def forward(self, noisy_img, step_i):
        # estimate the noise standard deviation from the noisy image
        sigma_est = np.mean(estimate_sigma(noisy_img, channel_axis=1))
        print(f'estimated noise standard deviation = {sigma_est}')

        patch_kw = dict(patch_size=self.patch_size(step_i),
                        patch_distance=self.patch_distance(step_i),  # 13x13 search area
                        channel_axis=1)
        denoised = denoise_nl_means(noisy_img, h=self.h(step_i) * sigma_est,
                                    fast_mode=self.fast_mode, **patch_kw)
        return denoised


class WNNMDenoiser(Denoiser):
    """
     Applies weighted nuclear norm minimization based denoising to the imput image img
     TODO: change channel dimension (now its on -1 implicitly)
    """

    def __init__(self, n_iters, patch_size, delta, c, K, sigma_n, N_threshold, N_iter=3):
        super(WNNMDenoiser, self).__init__()
        self.patch_size = self.init_p(patch_size, n_iters)
        self.delta = self.init_p(delta, n_iters)
        self.c = self.init_p(c, n_iters)
        self.K = self.init_p(K, n_iters)
        self.sigma_n = self.init_p(sigma_n, n_iters)
        self.N_threshold = self.init_p(N_threshold, n_iters)
        self.N_iter = self.init_p(N_iter, n_iters)  # the number of iterations for estimating \hat{X}_j

    def forward(self, noisy_img, step_i):
        patch_size = self.patch_size(step_i)

        # Specify the search window
        searchWindowRadius = patch_size * 3

        # Specify the width of padding and pad the noisy image
        pad = searchWindowRadius + patch_size
        imgPad = np.pad(noisy_img, pad_width=pad)
        imgPad = imgPad[..., pad:-pad]

        # Initialize variables to be iterated over
        xhat_iter = noisy_img

        for n in range(self.K(step_i)):
            # Pad the image for the iteration
            xhat_iter = np.pad(xhat_iter, pad_width=pad)
            xhat_iter = xhat_iter[..., pad:-pad]

            # Regularize the image that is denoised during the iteration
            y_iter = xhat_iter + self.delta(step_i) * (imgPad - xhat_iter)

            # Initialize the matrix to keep track of how many times each pixel has been updated
            pixel_contribution_matrix = np.ones_like(imgPad)

            # Identify similar patches and produce the matrix of similar patches
            for j in range(noisy_img.shape[0]):
                for i in range(noisy_img.shape[1]):
                    # Select the central patch
                    centerPatch = \
                        y_iter[j + searchWindowRadius:j + searchWindowRadius + 2 * patch_size,
                        i + searchWindowRadius:i + searchWindowRadius + 2 * patch_size:]

                    # Initialize the vector of distances between patches
                    dists = np.ones(((2 * searchWindowRadius + 1) ** 2))
                    # Initialize the matrix of patches
                    patches = np.zeros(((2 * searchWindowRadius + 1) ** 2, (2 * patch_size) ** 2))
                    # Compute distances between patches
                    # This is partially vectorized by using indexing to take out patches in a sliding window fashing
                    # out of a vertical slice through the search window
                    for k in range(2 * searchWindowRadius + 1):
                        # Take a vertical slice in the search window
                        otherPatch = y_iter[j:j + 2 * pad,
                                     i + k:i + k + 2 * patch_size,
                                     :]

                        # Determine indices corresponding to patches in a window sliding down the search window
                        indexer = np.arange((2 * patch_size) ** 2)[None, :] + (2 * patch_size) * np.arange(
                            otherPatch.shape[0] - 2 * patch_size + 1)[:, None]

                        # Set columns to be patches
                        otherPatch = otherPatch.flatten()
                        otherPatch = np.reshape(otherPatch[indexer],
                                                (otherPatch[indexer].shape[0], (2 * patch_size) ** 2))

                        # Compute distance and store the corresponding patches
                        dists[k * (2 * searchWindowRadius + 1):(k + 1) * (2 * searchWindowRadius + 1)] = (
                                np.sum((centerPatch.reshape(((2 * patch_size) ** 2)) - otherPatch) ** 2,
                                       axis=1) / (2 * patch_size) ** 2).flatten()
                        patches[k * (2 * searchWindowRadius + 1):(k + 1) * (2 * searchWindowRadius + 1), :] = otherPatch

                    # Select to N_threshold nearest patches and creat a patch matrix
                    indcs = np.argsort(dists)
                    Yj = (patches[indcs[:self.N_threshold(step_i)], :]).transpose()

                    # Center the columns
                    Yj_means = np.sum(Yj, axis=0)
                    Yj_center = Yj - Yj_means

                    # First iteration need to estimate singular values of Xj
                    U, S, V_T = svd(Yj_center, full_matrices=False)
                    sing_val = np.sqrt(np.maximum(S ** 2 - self.N_threshold(step_i) * self.sigma_n(step_i) ** 2, 0))

                    # Calculate the weights and sinfular values of \hat{X}_j iteratively
                    for m in range(self.N_iter(step_i)):
                        w = self.c(step_i) * np.sqrt(self.N_threshold(step_i)) / (sing_val + 10 ** (-6))
                        sing_val = np.diag(np.maximum(S - w, 0))

                    # Compute \hat{X}_j
                    Xj_hat_center = U @ np.diag(np.maximum(S - w, 0)) @ V_T
                    Xj_hat = Xj_hat_center + Yj_means

                    # Add the estimate of denoised central patch (first column of \hat{X}_j)
                    # to the esmated denoised image clipping it to between 0 and 1
                    xhat_iter[j + searchWindowRadius:j + searchWindowRadius + 2 * patch_size,
                    i + searchWindowRadius:i + searchWindowRadius + 2 * patch_size,
                    :] = xhat_iter[j + searchWindowRadius:j + searchWindowRadius + 2 * patch_size,
                         i + searchWindowRadius:i + searchWindowRadius + 2 * patch_size,
                         :] + np.clip(Xj_hat[:, 0].reshape((2 * patch_size, 2 * patch_size, 1)), 0, 1)

                    # Keep track of how many times each pixel has been added to
                    pixel_contribution_matrix[j + searchWindowRadius:j + searchWindowRadius + 2 * patch_size,
                    i + searchWindowRadius:i + searchWindowRadius + 2 * patch_size,
                    :] = pixel_contribution_matrix[j + searchWindowRadius:j + searchWindowRadius + 2 * patch_size,
                         i + searchWindowRadius:i + searchWindowRadius + 2 * patch_size,
                         :] + np.ones_like(
                        pixel_contribution_matrix[j + searchWindowRadius:j + searchWindowRadius + 2 * patch_size,
                        i + searchWindowRadius:i + searchWindowRadius + 2 * patch_size, :])

            # Remove the padding and average out contributions to pixels from different patches
            xhat_iter = xhat_iter[pad:-pad, pad:-pad, :] / pixel_contribution_matrix[pad:-pad, pad:-pad, :]

        # Produce the final output
        out = xhat_iter
        return out


class BM3DDenoise(Denoiser):
    def __init__(self, n_iters, sigma_psd=30 / 255):
        super(BM3DDenoise, self).__init__()
        self.sigma_psd = self.init_p(sigma_psd, n_iters)

    def forward(self, noisy_img, step_i):
        import bm3d
        noisy_img = noisy_img.clone().detach().cpu().numpy()
        clean_img = []
        for img in noisy_img:
            denoised_img = bm3d.bm3d(img,
                                     sigma_psd=self.sigma_psd(step_i),
                                     stage_arg=bm3d.BM3DStages.HARD_THRESHOLDING)
            clean_img.append(denoised_img)
        clean_img = np.stack(clean_img)
        return clean_img



import numpy as np
from scipy import ndimage, signal
from tqdm import tqdm
from ..tools.base_transforms import adjust_img_stats


def create_gabor(
        height=36,
        width=64,
        phase=0,
        wavelength=10,
        orientation=0,
        sigma=5,
        dy=0,
        dx=0,
        target_mean=None,
        target_contrast=None,
        img_min=-1,
        img_max=1,
):
    """
    :param height: Height of the image in pixels.
    :param width: Width of the image in pixels.
    :param phase: Angle at which to start the sinusoid in degrees.
    :param wavelength: Wavelength of the sinusoid (1 / spatial frequency) in pixels.
    :param orientation: Counterclockwise rotation to apply (0 is horizontal) in
    degrees.
    :param sigma: Sigma of the gaussian mask used in pixels
    :param dy: Amount of translation in y (positive moves down) in pixels/height.
    :param dx: Amount of translation in x (positive moves right) in pixels/height.
    :param target_mean: Target mean of the image. If None, no normalization is applied.
    :param target_contrast: Target contrast of the image. If None, no normalization is applied.
    :param img_min: Minimum value of the image.
    :param img_max: Maximum value of the image.

    :return: Array of height x width shape with the required gabor.
    """
    # Compute image size to avoid translation or rotation producing black spaces
    padding = max(height, width)
    imheight = height + 2 * padding
    imwidth = width + 2 * padding
    # we could have diff pad sizes per dimension = max(dim_size, sqrt((h/2)^2 + (w/2)^2))
    # but this simplifies the code for just a bit of inefficiency

    # Create sinusoid with right wavelength and phase
    start_sample = phase
    step_size = 360 / wavelength
    samples = start_sample + step_size * np.arange(imheight)
    samples = np.mod(samples, 360)  # in degrees
    rad_samples = samples * (np.pi / 180)  # radians
    sin = np.sin(rad_samples)

    # Create Gabor by stacking the sinusoid along the cols
    gabor = np.tile(sin, (imwidth, 1)).T

    # Rotate around center
    gabor = ndimage.rotate(gabor, orientation, reshape=False)

    # Apply gaussian mask
    gaussy = signal.gaussian(imheight, std=sigma)
    gaussx = signal.gaussian(imwidth, std=sigma)
    mask = np.outer(gaussy, gaussx)
    gabor = gabor * mask

    # Translate (this is only approximate but it should be good enough)
    if abs(dx) > 1 or abs(dy) > 1:
        raise ValueError('Please express translations as factors of the height/width,'
                         'i.e, a number in interval [-1, 1] ')
    dy = int(dy * height)  # int is the approximation
    dx = int(dx * width)
    gabor = gabor[padding - dy: -padding - dy, padding - dx: -padding - dx]

    if gabor.shape != (height, width):
        raise ValueError('Dimensions of gabor do not match desired dimensions.')

    if target_mean is not None and target_contrast is not None:
        gabor, _ = adjust_img_stats(gabor, mu=target_mean, sigma=target_contrast, img_min=img_min, img_max=img_max)
    elif target_mean is not None or target_contrast is not None:
        raise ValueError('If you want to adjust the mean or contrast, you must specify both.')

    return gabor.astype(np.float32)


def create_gabor_loader(param_ranges, load_path=None, save_path=None):
    """
    Grid search over gabor parameters then return a loader with these
    :param param_ranges: The ranges for the grid search
    :param load_path: Loading path the samples (new samples won't be created)
    :param save_path: The created samples will be saved here
    :return: Dataset with gabor filters
    """
    gabors = []

    if load_path is not None:
        import pickle
        with open(load_path, 'rb') as file:
            gabors = pickle.load(file)
        return gabors

    import itertools
    param_combinations = itertools.product(*list(param_ranges.values()))
    print(f'Creating gabors')
    for i, params in tqdm(enumerate(param_combinations)):
        param_values = dict(zip(param_ranges.keys(), params))
        gabor = create_gabor(**param_values)
        gabors.append(dict(image=gabor, params=param_values))

    if save_path is not None:
        import pickle
        with open(save_path, 'wb') as file:
            pickle.dump(gabors, file)

    return gabors



"""
Gabor parameters for grid search

height:         int         # (px) image height
width:          int         # (px) image width
phases:         long        # (degree) angle at which to start the sinusoid
wavelengths:    long        # (px) wavelength of the sinusoid (1 / spatial frequency)
orientations:   long        # (degree) counterclockwise rotation to apply (0 is horizontal, 90 vertical)
sigmas:         long        # (px) sigma of the gaussian mask used
dys:            long        # (px/height) amount of translation in y (positive moves downwards)
dxs:            long        # (px/width) amount of translation in x (positive moves right)
"""
default_gabor_ranges = \
    dict(
        height=[32],
        width=[32],
        phase=[0, 90, 180, 270],
        wavelength=[4, 7, 10, 15, 20],
        orientation=np.linspace(0, 180, 8, endpoint=False),
        sigma=[2, 3, 5, 7, 9],
        dy=np.linspace(-0.3, 0.3, 7),
        dx=np.linspace(-0.3, 0.3, 13)
    )

""" 
Limits of some parameters search range to find the optimal Gabor

height:             int         # (px) height of image 
width:              int         # (px) width of image
lower_phase:        float
upper_phase:       float
lower_wavelength:   float
upper_wavelength:   float
lower_orientation:  float
upper_orientation:  float
lower_sigma:        float
upper_sigma:        float
lower_dy:           float
upper_dy:           float
lower_dx:           float
upper_dx:           float
"""
default_gabor_limits = \
    [
        (0, 360),  # phase
        (4, 20),  # wavelength
        (0, 180),  # orientation
        (2, 9),  # sigma
        (-0.35, 0.35),  # dy
        (-0.35, 0.35)  # dx
    ]


import numpy as np
from torch.nn import Module, Sequential


def write_image_to_disk(filepath, image):
    from PIL import Image

    if image.shape[0] == 3:
        image = np.round(image * 127.5 + 127.5)
        image = image.astype(np.uint8)
        image = np.transpose(image, (1, 2, 0))
        im = Image.fromarray(image)
        im.save(filepath, format='png')
    else:
        for im in image:
            im = im.astype(np.uint8)
            while len(im.shape) > 2:
                im = np.squeeze(im, axis=0)
            im = Image.fromarray(im)
            im.save(filepath, format='png')


class SerializableSequential(Sequential):

    def __init__(self, *args):
        super().__init__(*args)

    def serialize(self):
        return [layer.serialize() for layer in self._modules.values()]

    @staticmethod
    def deserialize(serialized):
        sequential = SerializableSequential(*[
            layer["type"].deserialize(layer)
            if isinstance(layer, dict)
            else SerializableSequential.deserialize(layer)
            for layer in serialized
        ])
        return sequential


class SerializableModule(Module):

    def __init__(self):
        super().__init__()

    def serialize(self):
        return dict(type=self.__class__, params=None)

    @staticmethod
    def deserialize(serialized):
        return serialized["type"]


def batch_mean(batch, keepdim=False):
    """ Compute mean for a batch of images. """
    mean = batch.view(len(batch), -1).mean(-1)
    if keepdim:
        mean = mean.view(len(batch), 1, 1, 1)
    return mean


def batch_std(batch, keepdim=False, unbiased=True):
    """ Compute std for a batch of images. """
    std = batch.view(len(batch), -1).std(-1, unbiased=unbiased)
    if keepdim:
        std = std.view(len(batch), 1, 1, 1)
    return std



import torch
from torch import nn


class GaussianMixtureModel(nn.Module):
    def __init__(self, num_components, input_dim, fixed_variance=False):
        super(GaussianMixtureModel, self).__init__()

        self.num_components = num_components
        self.input_dim = input_dim

        # Parameters for the Gaussian mixture components
        self.weights = nn.Parameter(torch.randn(num_components))
        self.means = nn.Parameter(torch.randn(num_components, input_dim))
        if fixed_variance:
            self.variances = nn.Parameter(torch.ones(num_components, input_dim)*fixed_variance)
        else:
            self.variances = nn.Parameter(torch.rand(num_components, input_dim))

    def forward(self, x):
        # Compute the Gaussian component probabilities
        log_probs = []
        for i in range(self.num_components):
            weight = self.weights[i]
            mean = self.means[i]
            var = self.variances[i]
            log_prob = torch.log(weight) + self._log_gaussian(x, mean, var)
            log_probs.append(log_prob)

        # Calculate the log-sum-exp of log probabilities
        log_probs = torch.stack(log_probs, dim=1)
        log_sum_exp = torch.logsumexp(log_probs, dim=1)

        return log_sum_exp

    @staticmethod
    def _log_gaussian(x, mean, var):
        # Calculate the log likelihood of a Gaussian distribution
        return -0.5 * torch.sum(((x - mean) / var)**2, dim=1) - 0.5 * torch.sum(torch.log(var))



import numpy as np
import torch
import torch.nn.functional as F
import torchvision.transforms as T
from tqdm import tqdm
from ..tools.base_transforms import roll
from ..tools.utils import batch_std
from ..tools.precond import fft_smooth

eps = 1e-12


def deepdraw(process, operation):
    """
    Generate an image by iteratively optimizing activity of net.

    :param process: MEI_result process object
    :param operation: The operation to optimize for
    :return: The optimized image
    """

    for i in tqdm(range(process.iter_n)):
        # Diversity for multiple images (Image based)
        if process.diverse:
            div_term = diversion_term(process, **process.diverse_params)
        else:
            div_term = 0

        make_step(process, operation, step_i=i, add_loss=div_term)

        if i % process.save_every == 0:
            image = process.get_image().data.clone().cpu().numpy()
            process.image_history.append(image)

    result_stats = get_result_stats(process, operation)

    return result_stats


def make_step(process, operation, step_i, add_loss=0):
    """
    Update meitorch in place making a gradient ascent step in the output of net.

    :param process: MEI_result process object
    :param operation: The operation to optimize for
    :param step_i: The step index
    :param add_loss: An additional term to add to the network activation before
                        calling backward on it. Usually, some regularization.
    """
    process.optimizer.zero_grad()
    inputs = process.get_samples()

    if process.scaler:
        scale = process.scaler(step_i)
        inputs = F.interpolate(inputs, scale_factor=scale, mode='bilinear', align_corners=True)
        inputs = T.CenterCrop(process.img_shape[-2:])(inputs)

    # apply jitter shift
    if process.jitter:
        ox, oy = np.random.randint(-process.jitter, process.jitter + 1, 2)  # use uniform distribution
        ox, oy = int(ox), int(oy)
        inputs.data = roll(roll(inputs.data, ox, -1), oy, -2)

    # normalize the image in backpropagatable manner
    img = inputs
    if process.train_norm and process.train_norm > 0.0:
        img_idx = batch_std(inputs.data) + eps > process.train_norm / process.scale  # images to update
        if img_idx.any():
            img = inputs.clone()  # avoids overwriting original image but lets gradient through
            img[img_idx] = ((img[img_idx] / (batch_std(img[img_idx], keepdim=True) +
                                             eps)) * (process.train_norm / process.scale))

    outputs = operation(img)
    loss = outputs["objective"].mean()
    process.update_loss_history(outputs)

    if loss < process.best_loss:
        process.best_loss = loss
        process.best_image = process.get_image()

    loss += add_loss
    loss.backward()
    
    if process.precond:
        for param in process.parameters():
            if param.requires_grad and param.grad is not None and len(param.grad.size()) >= 2:
                smooth_grad = fft_smooth(param.grad, process.precond(step_i))
                if len(smooth_grad.size()) != len(param.grad.data.size()):
                    smooth_grad = smooth_grad.squeeze(0)
                param.grad.data.copy_(smooth_grad)

    process.optimizer.step()
    if process.scheduler is not None:
        process.scheduler.step()

    if process.norm and process.norm > 0.0:
        data_idx = batch_std(inputs.data) + eps > process.norm / process.scale
        inputs.data[data_idx] = (inputs.data / (batch_std(inputs.data, keepdim=True) + eps) * process.norm / process.scale)[
            data_idx]

    if process.jitter:
        # undo the shift
        inputs.data = roll(roll(inputs.data, -ox, -1), -oy, -2)

    if process.clip:
        inputs.data = torch.clamp(inputs.data, process.bias-process.scale, process.bias+process.scale)

    if process.blur:
        inputs.data = process.blur(inputs.data, step_i)


def get_result_stats(process, operation):
    with torch.no_grad():
        img = process.get_image()
        losses = operation(img)
        activation = losses["activation"].data.cpu().numpy()

    #cont, vals, lim_contrast = Analyze.contrast_tuning(
    #    operation, img.detach().cpu().numpy(), device=process.device)
    result_dict = dict(
        activation=activation,
    #    monotonic=bool(np.all(np.diff(vals) >= 0)),
    #    max_activation=np.max(vals),
    #    max_contrast=cont[np.argmax(vals)],
    #    sat_contrast=np.max(cont),
        img_mean=img.mean(),
        img_std=img.std()
    #    lim_contrast=lim_contrast,
    )

    return result_dict


def diversion_term(process, div_metric='correlation', div_linkage='minimum', div_weight=0, div_mask=1):
    """
    Compute the diversity term for a set of images.

    :param div_metric: What metric to use when computing pairwise differences.
    :param div_linkage: How to agglomerate pairwise distances.
    :param div_weight: Weight given to the diversity term in the objective function.
    :param div_mask: Array (height x width) used to mask irrelevant parts of the
                        image before calculating diversity.
    """
    img = process.get_image()
    mask = torch.tensor(div_mask, dtype=torch.float32, device=process.device)
    div_term = 0
    if div_weight > 0:
        # Compute distance matrix
        images = (img * mask).view(len(img), -1)  # num_images x num_pixels
        if div_metric == 'correlation':
            # computations restricted to the mask
            means = (images.sum(dim=-1) / mask.sum()).view(len(images), 1, 1, 1)
            residuals = ((img - means) * torch.sqrt(mask)).view(len(img), -1)
            ssr = (((img - means) ** 2) * mask).sum(-1).sum(-1).sum(-1)
            distance_matrix = -(torch.mm(residuals, residuals.t()) /
                                torch.sqrt(torch.ger(ssr, ssr)) + 1e-12)
        elif div_metric == 'cosine':
            image_norms = torch.norm(images, dim=-1)
            distance_matrix = -(torch.mm(images, images.t()) /
                                (torch.ger(image_norms, image_norms) + 1e-12))
        elif div_metric == 'euclidean':
            distance_matrix = torch.norm(images.unsqueeze(0) -
                                         images.unsqueeze(1), dim=-1)
        else:
            raise ValueError('Invalid distance metric {} for the diversity term'.format(div_metric))

        # Compute overall distance in this image set
        triu_idx = torch.triu(torch.ones(len(distance_matrix),
                                         len(distance_matrix)), diagonal=1) == 1
        if div_linkage == 'minimum':
            distance = distance_matrix[triu_idx].min()
        elif div_linkage == 'average':
            distance = distance_matrix[triu_idx].mean()
        else:
            raise ValueError('Invalid linkage for the diversity term: {}'.format(div_linkage))

        div_term = div_weight * distance
    return div_term



from scipy.ndimage.filters import gaussian_filter
from .base_transforms import remove_small_area
import numpy as np
import torch
from .base_transforms import fit_gauss_envelope
from skimage.morphology import convex_hull_image, binary_erosion
from scipy.ndimage import binary_erosion, generate_binary_structure


def gaussian_mask(img, factor):
    """
    Create a gaussian mask based on the image

    :param img: The image
    :param factor: Strength of the mask
    :return: The mask
    """
    *_, mask = fit_gauss_envelope(img)
    return mask ** factor


def gaussian_mask_with_info(img, factor):
    """
    Create a gaussian mask based on the image

    :param img: The image
    :param factor: Strength of the mask
    :return: The mask, the mean, the covariance matrix
    """
    mu, cov, mask = fit_gauss_envelope(img)
    mu_x = mu[0]
    mu_y = mu[1]
    cov_x = cov[0, 0]
    cov_y = cov[1, 1]
    cov_xy = cov[0, 1]
    mei_mask = mask ** factor
    return mei_mask, mu_x, mu_y, cov_x, cov_y, cov_xy


def mei_mask(img, delta_thr=0.5, size_thr=50, expansion_sigma=3, expansion_thr=0.3, filter_sigma=2):
    """
    Create a mask for the MEI image

    :param img: The image
    :param delta_thr: The threshold for the delta
    :param size_thr: The threshold for the size
    :param expansion_sigma: The sigma for the expansion
    :param expansion_thr: The threshold for the expansion
    :param filter_sigma: The sigma for the filter
    :return: The mask
    """
    img = img.squeeze()
    delta = img - img.mean()
    mask = np.abs(delta) > delta_thr
    # remove small lobes - likely an artifact
    mask = remove_small_area(mask, size_threshold=size_thr)
    # fill in the gap between lobes
    mask = convex_hull_image(mask)
    # expand the size of the mask
    mask = gaussian_filter(mask.astype(float), sigma=expansion_sigma) > expansion_thr
    # blur the edge, giving smooth transition
    mask = gaussian_filter(mask.astype(float), sigma=filter_sigma)
    return mask


def mei_tight_mask(img, operation, device, stdev_size_thr=1, filter_sigma=1, target_reduction_ratio=0.9):
    """
    Create a tight mask for the MEI image

    :param img: The image
    :param operation: The operation of the MEI
    :param device: The device
    :param stdev_size_thr: The threshold for the standard deviation
    :param filter_sigma: The sigma for the filter
    :param target_reduction_ratio: The target reduction ratio
    :return: The mask, the reduction ratio
    """
    def get_activation(mei):
        with torch.no_grad():
            img = torch.Tensor(mei).to(device)
            activation = operation(img).data.cpu().numpy()
        return activation

    img = img.squeeze()
    delta = img - img.mean()
    fluc = np.abs(delta)
    thr = np.std(fluc) * stdev_size_thr

    # original mask
    mask = convex_hull_image((fluc > thr).astype(float))
    fm = gaussian_filter(mask.astype(float), sigma=filter_sigma)
    masked_img = fm * img + (1 - fm) * img.mean()
    activation = base_line = get_activation(masked_img)

    # TODO: fix for multiple channels
    count = 0
    while activation > base_line * target_reduction_ratio:
        selem_size = 3
        selem = generate_binary_structure(img.ndim, 1)
        selem[selem_size//2] = 1
        mask = binary_erosion(mask, selem)
        fm = gaussian_filter(mask.astype(float), sigma=filter_sigma)
        masked_img = fm * img + (1 - fm) * img.mean()
        activation = get_activation(masked_img)
        count += 1

        if count > 100:
            print('This has been going on for too long! - aborting')
            raise ValueError('The activation does not reduce for the given setting')

    reduction_ratio = activation / base_line
    return fm, reduction_ratio


def mask_image(img, mask='gaussian', background=0, operation=lambda x: x, device='cpu', **MaskParams):
    """
    Applies the mask `mask` onto the `img`. The completely masked area is then
    replaced with the value `background`.

    :param img: image to be masked
    :param mask: type of mask to be applied
    :param background: value to be used for the masked area
    :param operation: operation to be applied on the masked image
    :param device: device to be used for the operation
    :param MaskParams: parameters for the mask
    :return: masked image
    """
    if mask is None:
        return img
    elif mask == 'gaussian':
        factor = MaskParams.get('factor', 1/4)
        _mask = gaussian_mask(img, factor)
    elif mask == 'meitorch':
        _mask = mei_mask(img, **MaskParams)
    elif mask == 'mei_tight':
        _mask, _ = mei_tight_mask(img, operation, device, **MaskParams)
    else:
        raise ValueError(f'Unknown mask type: {mask}')

    filler = np.full_like(img, background)
    return img * _mask + filler * (1 - _mask)



import torch
import numpy as np


def fft_smooth(grad, factor=1/4):
    """
    Tones down the gradient with 1/sqrt(f) filter in the Fourier domain.
    Equivalent to low-pass filtering in the spatial domain.



    :param grad: The gradient
    :param factor: The factor
    """
    if factor == 0:
        return grad
    h, w = grad.size()[-2:]
    tw = np.minimum(np.arange(0, w), np.arange(w, 0, -1), dtype=np.float32)
    th = np.minimum(np.arange(0, h), np.arange(h, 0, -1), dtype=np.float32)
    t = 1 / np.maximum(1.0, (tw[None, :] ** 2 + th[:, None] ** 2) ** factor)
    F = grad.new_tensor(t / t.mean()).unsqueeze(0)
    pp = torch.fft.fft2(grad.data, dim=(-1, -2))
    return torch.real(torch.fft.ifft2(pp * F, dim=(-1, -2)))



import torch
import numpy as np
from tqdm import tqdm
from itertools import product

from .tools.base_transforms import roll
from .tools.masks import mask_image


class Analyze:

    @staticmethod
    def best_match(mei_result, operation, dataloader, mask=None, **MaskParams):
        """
        Find the image that maximizes the activation of the operation

        :param mei_result: The result object containing the optimized visualization
        :param operation: The optimization operation
        :param dataloader: The dataset containing the images
        :param mask: The mask to be applied to the image (optional)
        :param MaskParams: The parameters of the specific mask
        :return: The activation and the image that maximizes the activation
        """
        img_activations = []
        for image, label in tqdm(dataloader.dataset):
            image = mask_image(image, mask, mei_result.bias, **MaskParams)
            if len(image.shape) == 2:
                image = np.expand_dims(image, axis=0)
            if type(image) is not torch.Tensor:
                image = torch.tensor(image, dtype=torch.float32, requires_grad=True, device=mei_result.device)
            else:
                image.requires_grad = True
            y = operation(image)
            img_activations.append(y.item())

        img_activations = np.array(img_activations)
        pos = np.argmax(img_activations)
        return img_activations[pos], dataloader.dataset[pos][0].squeeze(0)

    @staticmethod
    def masked(mei_result, mask, **MaskParams):
        """
        Apply a mask to the image

        :param mei_result: The result object containing the optimized visualization
        :param mask: The mask to be applied (gaussian, meitorch, tight_mei)
        :param MaskParams: The parameters for the specific mask
        :return: The masked image
        """
        if mei_result.image is None:
            return None
        return mask_image(mei_result.image, mask, mei_result.bias, **MaskParams)

    @staticmethod
    def masked_responses_process(mei_result, operation=None, mask='gaussian', **MaskParams):
        """
        Apply a mask to the image and return the activation of the operation

        :param mei_result: The result object containing the optimized visualization
        :param operation: The operation on which the activation will be measured
        :param mask: The mask to be applied (gaussian, meitorch, tight_mei)
        :param MaskParams: The parameters for the specific mask
        :return: The activation and the masked image
        """
        if mei_result.image is None:
            return None
        _, masked_img_activations, masked_images = \
            Analyze.masked_responses([mei_result.image], operation, mask, mei_result.bias, **MaskParams)
        return masked_img_activations[0], masked_images[0]

    @staticmethod
    def jittered_responses(mei_result, operation, jitter_size):
        """
        Jitter the image and return the activation of the operation

        :param mei_result: The result object containing the optimized visualization
        :param operation: The operation on which the activation will be measured
        :param jitter_size: The size of the jitter,
                            the image will be shifted by -jitter_size, ..., 0, ..., jitter_size
        :return: The activation and the jittered images
        """
        if mei_result.image is None:
            return None

        # jitter_size = 0 vagy 5
        shift = list(enumerate(range(-jitter_size, jitter_size + 1)))
        activations = np.empty((len(shift), len(shift)))

        jiterred_images = []

        with torch.no_grad():
            img = torch.Tensor(mei_result.image).to(mei_result.device)

            for (iy, jitter_y), (ix, jitter_x) in product(shift, shift):
                jitter_y, jitter_x = int(jitter_y), int(jitter_x)
                jittered_img = roll(roll(img, jitter_y, -2), jitter_x, -1)
                jiterred_images.append(jittered_img)
                activations[iy, ix] = operation(jittered_img).data.cpu().numpy()

        return activations, jiterred_images

    @staticmethod
    def shifted_response(mei_result, operation, x_shift, y_shift):
        """
        Shift the image and return the activation of the operation

        :param mei_result: The result object containing the optimized visualization
        :param operation: The operation on which the activation will be measured
        :param x_shift: Shift on the x-axis
        :param y_shift: Shift on the y-axis
        :return: The activation and the shifted image
        """

        if mei_result.image is None:
            return None

        shifted_mei = np.roll(np.roll(mei_result.image, x_shift, 1), y_shift, 0)

        with torch.no_grad():
            shifted_mei = torch.Tensor(shifted_mei).to(mei_result.device)
            activations = operation(shifted_mei).data.cpu().numpy()

        return activations, shifted_mei

    @staticmethod
    def masked_responses(images, operation, mask='gaussian', bias=0, device='cpu', **MaskParams):
        """
        For comparision of activision between original and masked samples

        :param images: the images to be masked
        :param operation: the operation on which activisions will be measured
        :param mask: the type of the mask to be applied
        :param bias: background color
        :param device: device
        :param MaskParams: the parameters of specific masks (more on this in utils)
        :return: activation of the original image, activation of the masked image, the masked image
        """

        def evaluate_image(x):
            if len(x.shape) == 2:
                x = np.expand_dims(x, axis=0)
            x = torch.tensor(x, dtype=torch.float32, requires_grad=False, device=device)
            y = operation(x).data.cpu().numpy()
            return y

        original_img_activations = []
        masked_img_activations = []
        masked_images = []

        for image in tqdm(images):
            original_img_activations.append(evaluate_image(image))
            masked_image = mask_image(image, mask, bias, operation=operation, **MaskParams)
            masked_images.append(masked_image)
            masked_img_activations.append(evaluate_image(masked_image))

        return original_img_activations, masked_img_activations, masked_images

    @staticmethod
    def compute_spatial_frequency(img):
        """
        Computes and plots the spatial frequency of a given image

        :param img: image to compute sf on
        :return: frequency of each column, row in arrays, magnitude spectrum
        """
        from matplotlib import pyplot as plt

        # Compute the 2D Fourier Transform
        fft_img = np.fft.fft2(img)

        # Shift the zero-frequency component to the center of the spectrum
        fft_img_shifted = np.fft.fftshift(fft_img)

        # Compute the magnitude spectrum (absolute value)
        magnitude_spectrum = np.abs(fft_img_shifted)

        if len(img.shape) == 2:
            rows, cols = img.shape
            plt.imshow(np.log1p(magnitude_spectrum), cmap='gray')
        else:
            channels, rows, cols = img.shape
            for channel in range(channels):
                plt.imshow(np.log1p(magnitude_spectrum[channel]), cmap='gray')

        plt.colorbar()
        plt.show()

        freq_rows = np.fft.fftfreq(rows)
        freq_cols = np.fft.fftfreq(cols)

        return freq_cols, freq_rows, magnitude_spectrum

    @staticmethod
    def contrast_tuning(model, img, min_contrast=0.01, n=1000, linear=True, use_max_lim=False, device='cpu'):
        """
        Computes the contrast tuning curve for the given image and model.

        :param model: The model
        :param img: The image to compute the tuning curve for
        :param min_contrast: The minimum contrast to use
        :param n: The number of points to use
        :param linear: Whether to use linearly spaced points
        :param use_max_lim: Whether to use the maximum possible contrast without clipping
        :param device: The device to use
        :return: The contrast tuning curve
        """
        mu = img.mean()
        delta = img - img.mean()
        vmax = delta.max()
        vmin = delta.min()

        min_pdist = delta[delta > 0].min()
        min_ndist = (-delta[delta < 0]).min()

        max_lim_gain = max((1 - mu) / min_pdist, mu / min_ndist)

        base_contrast = img.std()

        lim_contrast = 1 / (vmax - vmin) * base_contrast  # maximum possible reachable contrast without clipping
        min_gain = min_contrast / base_contrast
        max_gain = min((1 - mu) / vmax, -mu / vmin)

        def run(x):
            with torch.no_grad():
                img = torch.Tensor(x).to(device)
                result = model(img)["objective"]
            return result

        target = max_lim_gain if use_max_lim else max_gain

        if linear:
            gains = np.linspace(min_gain, target, n)
        else:
            gains = np.logspace(np.log10(min_gain), np.log10(target), n)
        vals = []
        cont = []

        for g in tqdm(gains):
            img = delta * g + mu
            img = np.clip(img, 0, 1)
            c = img.std()
            v = run(img).data.cpu().numpy()
            cont.append(c)
            vals.append(v)

        vals = np.array(vals)
        cont = np.array(cont)

        return cont, vals, lim_contrast



from abc import ABC, abstractmethod
from torch import optim
import numpy as np


def LinearSchedule(start, end):
    return lambda iter_n: _LinearSchedule(start, end, iter_n)


def OctaveSchedule(values: list):
    return lambda iter_n: _OctaveSchedule(values, iter_n)


def ConstantSchedule(value):
    return lambda iter_n: _ConstantSchedule(value)


def RandomSchedule(minimum, maximum):
    return lambda iter_n: _RandomSchedule(minimum, maximum)


class Scheduler(ABC):
    @abstractmethod
    def __call__(self, step):
        pass
    
    def __str__(self):
        pass


class _LinearSchedule(Scheduler):
    def __init__(self, start, end, steps):
        self.start = start
        self.end = end
        self.steps = steps

    def __call__(self, step):
        return self.start + (self.end - self.start) * step / self.steps
    
    def __str__(self):
        return f"linear {self.start}->{self.end}"
    
    
class _RandomSchedule(Scheduler):
    def __init__(self, minimum, maximum):
        self.maximum = maximum
        self.minimum = minimum
        
    def __call__(self, step):
        return np.random.uniform(self.minimum, self.maximum)
    
    def __str__(self):
        return f"random {self.minimum}->{self.maximum}"


class _OctaveSchedule(Scheduler):
    def __init__(self, values: list, steps):
        self.values = values
        self.steps = steps

    def __call__(self, step):
        octaves = len(self.values)
        steps_per_octave = self.steps // octaves + 1
        step = step // steps_per_octave
        return self.values[step]
    
    def __str__(self):
        return f"octave {self.values}"


class _ConstantSchedule(Scheduler):
    def __init__(self, value):
        self.value = value

    def __call__(self, step):
        return self.value
    
    def __str__(self):
        return f"constant {self.value}"


def get_lr_scheduler(optimizer, hparams):

    """
    StepLR:     This scheduler decreases the learning rate at specified epochs by a multiplicative factor.
                For example, you could use this scheduler to decrease the learning rate by 0.5 every 10 epochs.

                - step_size (int): The number of epochs between each learning rate decrease.
                - gamma (float): The multiplicative factor by which the learning rate is decreased at each step.

    MultiStepLR: This scheduler is similar to StepLR, but it allows you to specify multiple epochs
                    at which to decrease the learning rate. For example,
                    you could use this scheduler to decrease the learning rate by 0.5 at epochs 10, 20, and 30.

                - milestones (list[int]): A list of epochs at which to decrease the learning rate.
                - gamma (float): The multiplicative factor by which the learning rate is decreased at each step.

    ExponentialLR: This scheduler decreases the learning rate exponentially by a multiplicative factor at each epoch.

                - gamma (float): The multiplicative factor by which the learning rate is decreased at each epoch.

    CosineAnnealingLR: This scheduler decreases the learning rate following a cosine curve,
                        starting from a high value and gradually decreasing to a low value.

                - T_max (int): The maximum number of epochs.
                - eta_min (float): The minimum learning rate.

    CosineAnnealingWarmRestartsLR: This scheduler is similar to CosineAnnealingLR,
                                    but it restarts the cosine annealing process at specified epochs.
                                    This can be useful for preventing the model from getting stuck in local minima.

                - T_max (int): The maximum number of epochs per restart.
                - eta_min (float): The minimum learning rate.
                - T_mult (float): A factor by which the T_max is multiplied after each restart.

    :param optimizer: the optimizer to be used
    :param hparams: the hyperparameters of the schdeler
    :return: the scheduler
    """

    s_type = hparams["type"]
    del hparams["type"]

    if s_type == "step":
        return optim.lr_scheduler.StepLR(optimizer, **hparams)
    elif s_type == "multi_step":
        return optim.lr_scheduler.MultiStepLR(optimizer, **hparams)
    elif s_type == "exponential":
        return optim.lr_scheduler.ExponentialLR(optimizer, **hparams)
    elif s_type == "cosine":
        return optim.lr_scheduler.CosineAnnealingLR(optimizer, **hparams)
    elif s_type == "cosine_warm":
        return optim.lr_scheduler.CosineAnnealingWarmRestarts(optimizer, **hparams)




from torch import nn
import numpy as np
import torch


class GenerativeMLPNet(nn.Module):
    """
    Parametric multilayer perceptron network

    :param hidden_sizes: list of int, the sizes of the hidden layers
    :param output_shape: int, the size of the output
    :param trainable_input: bool, whether the input of the transformation should be trainable
    :param activation: torch.nn.Module, the activation function to use
    :param activate_output: bool, whether to apply the activation function to the output
    :param distribution_base: str, the base distribution to use (normal or laplace)
    :param fixed_stddev: float, the fixed standard deviation to use for the output, False if it should be learned
    """

    def __init__(self, hidden_sizes, output_shape, trainable_input=False,
                 activation=nn.ReLU(), activate_output=True, distribution_base='normal', fixed_stddev=0.4):
        super(GenerativeMLPNet, self).__init__()
        self.hidden_sizes = hidden_sizes
        self.activation = activation
        self.activate_output = activate_output
        self.output_shape = output_shape
        self.fixed_stddev = fixed_stddev
        self.input = nn.Parameter(torch.randn(hidden_sizes[0]),
                                  requires_grad=trainable_input)

        layers = []
        output_shape = np.prod(output_shape) if fixed_stddev else 2*np.prod(output_shape)
        sizes = hidden_sizes + [output_shape]
        for i in range(len(sizes) - 1):
            layers.append(nn.Linear(sizes[i], sizes[i + 1]))
            if i < len(sizes) - 2 or activate_output:
                layers.append(self.activation)
        self.mlp_layers = nn.Sequential(*layers)
        
        self.distribution = distribution_base
        if distribution_base == 'normal':
            self.dist = torch.distributions.Normal
        elif distribution_base == 'laplace':
            self.dist = torch.distributions.Laplace
        else:
            raise ValueError('Unknown distribution base: {}'.format(distribution_base))

    def forward(self, batch_size, use_mean=False):
        batch = torch.tile(self.input, (batch_size, 1))
        output = self.mlp_layers(batch)
        if self.fixed_stddev:
            y_mu = output
            y_sigma = torch.ones_like(output) * self.fixed_stddev
        else:
            y_mu, y_sigma = output.reshape(batch_size, 2, -1).chunk(2, dim=1)
        output_distribution = self.dist(y_mu, torch.exp(y_sigma))
        samples = output_distribution.sample() if not use_mean else output_distribution.mean
        return samples.reshape(batch_size, *self.output_shape)
    
    def __str__(self):
        return f" mlp net | hidden_sizes: {self.hidden_sizes} | " \
               f"distribution: {self.distribution} | fixed_stddev: {self.fixed_stddev} "


class GenerativeConvNet(nn.Module):
    """
    Parametric convolutional network

    :param hidden_sizes: list of int, the sizes of the hidden layers
    :param shape: tuple of int, the shape of the output
    :param kernel_size: int, the size of the convolutional kernel
    :param activation: torch.nn.Module, the activation function to use
    :param activate_output: bool, whether to apply the activation function to the output
    :param distribution_base: str, the base distribution to use (normal or laplace)
    :param fixed_stddev: float, the fixed standard deviation to use for the output, False if it should be learned
    :param use_mean: bool, whether to use the mean of the distribution instead of sampling
    """

    def __init__(self, hidden_sizes, shape, kernel_size, activation=nn.SiLU(), activate_output=False,
                 distribution_base='normal', fixed_stddev=0.4, use_mean=False, scale=1):
        super(GenerativeConvNet, self).__init__()
        self.shape = shape
        self.hidden_sizes = hidden_sizes
        self.activation = activation
        self.activate_output = activate_output
        self.scale = scale
        inputs = torch.tensor(self.generate_random_noise(self.shape), dtype=torch.float32)
        self.input = nn.Parameter(inputs, requires_grad=True)
        self.fixed_stddev = fixed_stddev
        self.kernel_size = kernel_size
        self.distribution_base = distribution_base
        self.use_mean = use_mean

        layers = []
        output_channels = shape[0] if fixed_stddev else 2*shape[0]
        sizes = hidden_sizes + [output_channels]
        if len(sizes) == 1:
            conv2d = nn.Conv2d(
                in_channels=sizes[0], out_channels=sizes[0],
                kernel_size=kernel_size, padding='same')
            layers.append(conv2d)
        else:
            for i in range(len(sizes) - 1):
                conv2d = nn.Conv2d(
                        in_channels=sizes[i], out_channels=sizes[i + 1],
                        kernel_size=kernel_size, padding='same')
                layers.append(conv2d)
                if i < len(sizes) - 2 or activate_output:
                    layers.append(self.activation)
        self.conv = nn.Sequential(*layers)

        if distribution_base == 'normal':
            self.dist = torch.distributions.Normal
        elif distribution_base == 'laplace':
            self.dist = torch.distributions.Laplace
        else:
            raise ValueError(f'Unknown distribution base: {distribution_base}')

    def forward(self, batch_size, use_mean=False):
        batch = torch.tile(self.input, (batch_size, 1, 1, 1))
        output = self.conv(batch)
        if self.fixed_stddev:
            y_mu = output
            y_sigma = torch.ones_like(output) * self.fixed_stddev
        else:
            y_mu, y_sigma = output.reshape(batch_size, 2, -1).chunk(2, dim=1)
        output_distribution = self.dist(y_mu, y_sigma)
        samples = output_distribution.mean if use_mean or self.use_mean else output_distribution.rsample()
        return samples.reshape(batch_size, *self.shape)

    def generate_random_noise(self, shape):
        # generate initial random image
        background_color = np.float32([0])
        gen_img = np.random.normal(background_color, self.scale/4, shape)
        gen_img = np.clip(gen_img, -1, 1)
        return gen_img

    def __str__(self):
        return f" cnn net | hidden_sizes: {self.hidden_sizes} | kernel_size: {self.kernel_size} | " \
               f"distribution: {self.distribution} | fixed_stddev: {self.fixed_stddev} "



import torch

from .linearmei import LinearMEI
from .result import MEI_image, MEI_variational, MEI_transformation
from .objective.deepdraw import deepdraw


class MEI(LinearMEI):
    """
    Class for generating more complex optimized inputs
    """
    def __init__(self, operation, shape=(1, 28, 28),  device='cpu'):
        super().__init__(operation, shape, device=device)

    def generate_pixel_mei(self, init=None, **MEIParams):
        """
        Generate most exciting inputs with pixel optimization
        :param init: the initial image (no random generation)
        :param MEIParams: Additional parameters for the optimization process.
        :return: MEI_image result
        """
        n_samples = MEIParams["n_samples"]
        del MEIParams["n_samples"]
        process = MEI_image(self.img_shape, n_samples, init=init, device=self.device, **MEIParams)
        return self._generate(process)

    def generate_variational_mei(self, init=None, **MEIParams):
        """
        Generate most exciting inputs with varitational optimization
        :param init: the initial image (no random generation)
        :param MEIParams: Additional parameters for the optimization process.
        :return: MEI_variational result
        """
        distribution = MEIParams["distribution"]
        del MEIParams["distribution"]
        process = MEI_variational(distribution, self.img_shape, init=init, device=self.device, **MEIParams)
        return self._generate(process)

    def generate_transformation_mei(self, **MEIParams):
        """
        Generate most exciting inputs with transformation optimization
        :param MEIParams: Additional parameters for the optimization process.
        :return: MEI_transformation result
        """
        transform = MEIParams["transform"]
        del MEIParams["transform"]
        process = MEI_transformation(transform, self.img_shape, device=self.device, **MEIParams)
        return self._generate(process)

    def _generate(self, process):
        if self.is_gradient_rf_op(process.param_dict):
            op, pointrf = self.to_gradient_rf_op(self.operation)
            process.result_dict.update({"pointrf": pointrf})
        else:
            op = self.operation
        result_dict = deepdraw(process, op)
        process.result_dict.update(result_dict)
        return process

    @staticmethod
    def is_gradient_rf_op(MEIParams):
        return "gradient_rf" in MEIParams and MEIParams["gradient_rf"]

    def to_gradient_rf_op(self, operation):
        """
        Generate most exciting inputs based on the linear function of the gradients of the input
        Uses deepdraw to optimize images
        :return: Process(es) with GradientRF images
        """

        X = torch.zeros(1, *self.img_shape, device=self.device, requires_grad=True)
        y = operation(X)
        y.backward()
        point_rf = X.grad.data.cpu().numpy().squeeze()
        rf = X.grad.data

        def linear_model(x):
            return (x * rf).sum()
        return linear_model, point_rf




from abc import ABC, abstractmethod
from torch import nn
import torch
import numpy as np

from .tools.denoisers import Denoiser
from .objective.optimizer import get_optimizer
from .tools.schedules import get_lr_scheduler


class MEI_result(nn.Module, ABC):
    def __init__(self, shape, n_samples=1, device='cpu',  **MEIParams):
        super(MEI_result, self).__init__()
        self.img_shape = shape
        self.n_samples = n_samples
        self.param_dict = MEIParams
        self.result_dict = dict()
        self.device = device
        self.scheduler = None
        self.optimizer = None

        assert "iter_n" in self.param_dict.keys(), "iter_n must be specified"
        self.iter_n = self.param_dict["iter_n"]

        self.save_every = self.param_dict["save_every"] if "save_every" in self.param_dict.keys() else self.iter_n

        if "scaler" in self.param_dict.keys() and self.param_dict["scaler"] is not None:
            if isinstance(self.param_dict["scaler"], (int, float)):
                self.scaler = lambda step: self.param_dict["scaler"]
            elif callable(self.param_dict["scaler"]):
                self.scaler = self.param_dict["scaler"](self.iter_n)
            else:
                raise ValueError("scaler must be float or Scheduler")
        else:
            self.scaler = None

        if "precond" in self.param_dict.keys() and self.param_dict["precond"] is not None:
            if isinstance(self.param_dict["precond"], (int, float)):
                self.precond = lambda step: self.param_dict["precond"]
            elif callable(self.param_dict["precond"]):
                self.precond = self.param_dict["precond"](self.iter_n)
            else:
                raise ValueError("precond must be float or Scheduler")
        else:
            self.precond = None

        if "jitter" in self.param_dict.keys() and self.param_dict["jitter"] is not None:
            self.jitter = self.param_dict["jitter"]
        else:
            self.jitter = None

        if "blur" in self.param_dict.keys() and self.param_dict["blur"] is not None:
            if isinstance(self.param_dict["blur"], Denoiser):
                self.blur = self.param_dict["blur"]
            elif isinstance(self.param_dict["blur"], str):
                assert "blur_params" in self.param_dict.keys(), "blur_params must be specified"
                self.blur = Denoiser.get_denoiser(self.param_dict["blur"],
                                                  self.iter_n,  **self.param_dict["blur_params"])
        else:
            self.blur = None

        if "clip" in self.param_dict.keys() and self.param_dict["clip"] is not None:
            self.clip = self.param_dict["clip"]
        else:
            self.clip = None

        if "norm" in self.param_dict.keys() and self.param_dict["norm"] is not None:
            self.norm = self.param_dict["norm"]
        else:
            self.norm = None

        if "train_norm" in self.param_dict.keys() and self.param_dict["train_norm"] is not None:
            self.train_norm = self.param_dict["train_norm"]
        else:
            self.train_norm = None

        self.diverse = "diverse" in self.param_dict.keys() and self.param_dict["diverse"]
        if self.diverse:
            if "diverse_params" in self.param_dict.keys():
                self.diverse_params = self.param_dict["diverse_params"]
            else:
                self.diverse_params = dict()

        self.best_loss = np.inf
        self.best_image = None

        assert "bias" in self.param_dict.keys(), "bias must be specified"
        self.bias = self.param_dict["bias"]
        assert "scale" in self.param_dict.keys(), "scale must be specified"
        self.scale = self.param_dict["scale"]

        self.loss_history = dict()
        self.image_history = []

    def show_results(self):
        for key, value in self.result_dict.items():
            print(key, value)

    def init_optimizer(self):
        assert "optimizer" in self.param_dict.keys(), "optimizer must be specified"
        assert "optimizer_params" in self.param_dict.keys(), "optimizer_hparams must be specified"

        optimizer_params = self.param_dict["optimizer_params"]
        scheduler_params = None
        if "scheduler_params" in optimizer_params.keys():
            scheduler_params = optimizer_params["scheduler_params"]
            del optimizer_params["scheduler_params"]

        self.optimizer = get_optimizer(self.parameters(),
                                       self.param_dict["optimizer"],
                                       self.iter_n, **optimizer_params)

        if scheduler_params is not None:
            self.scheduler = get_lr_scheduler(self.optimizer, scheduler_params)

    def spatial_frequency(self):
        """
        Compute the spatial frequency of the image and plot it
        :return: frequency of each column, row in arrays, magnitude spectrum
        """
        from .analyze import Analyze
        return Analyze.compute_spatial_frequency(self.get_image()[0])

    def update_loss_history(self, loss_dict):
        for key, value in loss_dict.items():
            if key not in self.loss_history.keys():
                self.loss_history[key] = []
            if isinstance(value, torch.Tensor):
                value = value.detach().cpu().numpy().mean()
            self.loss_history[key].append(value)

    def plot_losses(self, show=False, save_path=None, ranges=None):
        # plot multiple losses in one plot in different colors
        import matplotlib.pyplot as plt 
        fig, ax = plt.subplots()
        for key, value in self.loss_history.items():
            if ranges is not None:
                assert len(ranges) == 2
                clipped_value = np.clip(value, *ranges)
            else:
                clipped_value = value
            ax.plot(clipped_value, label=key)
        ax.legend()
        ax.set_ylabel("iteration")
        if show:
            plt.show()
        if save_path is not None:
            plt.savefig(save_path)
        return fig

    def plot_image_and_losses(self, save_path=None, ranges=None):
        # plot multiple losses in one plot in different colors
        import matplotlib.pyplot as plt 
        fig, (ax_im, ax_loss) = plt.subplots(1, 2, figsize=(6, 3),
                                             gridspec_kw={"width_ratios": [1, 1], "height_ratios": [1]})
        image = self.get_image()[0].detach().cpu().numpy()
        ax_im.imshow(image.reshape(40, 40))
        ax_im.set_title(f"Activation: {self.get_activation():.2f}")
        ax_im.axis("off")

        for key, value in self.loss_history.items():
            if ranges is not None:
                assert len(ranges) == 2
                clipped_value = np.clip(value, *ranges)
            else:
                clipped_value = value
            ax_loss.plot(clipped_value, label=key)
        ax_loss.legend()
        
        fig.tight_layout()

        if save_path is not None:
            plt.savefig(save_path)
        return fig

    @abstractmethod
    def get_image(self):
        pass

    @abstractmethod
    def get_samples(self):
        pass

    def show_image(self):
        import matplotlib.pyplot as plt
        image = self.get_image().detach().cpu().numpy()
        if image is None:
            return None
        if image.shape[0] == 1:
            image = image.squeeze(0)
        plt.imshow(image)
        plt.show()

    def save_image(self, path):
        from meitorch.tools.utils import write_image_to_disk
        image = self.get_image()
        if image is None:
            return None
        write_image_to_disk(path, image.detach().cpu().numpy())

    @abstractmethod
    def __getstate__(self):
        state = dict(
            img_shape=self.img_shape,
            n_samples=self.n_samples,
            param_dict=self.param_dict,
            result_dict=self.result_dict,
        )
        return state

    @abstractmethod
    def __setstate__(self, state):
        self.img_shape = state["img_shape"]
        self.n_samples = state["n_samples"]
        self.param_dict = state["param_dict"]
        self.result_dict = state["result_dict"]

    def get_activation(self):
        if "activation" not in self.result_dict.keys():
            return None
        return self.result_dict["activation"]

    def generate_random_noise(self, shape):
        # generate initial random image
        background_color = np.float32([self.bias] * self.img_shape[-1])
        gen_img = np.random.normal(background_color, self.scale / 16, shape)
        gen_img = np.clip(gen_img, -1, 1)
        return gen_img


class MEI_image(MEI_result):
    def __init__(self, shape, n_samples=1, init=None, device='cpu', **MEIParams):
        super(MEI_image, self).__init__(shape, n_samples, device=device, **MEIParams)
        self.batch_shape = (n_samples, *shape)

        if init is None:
            self.image = torch.nn.Parameter(
                torch.tensor(self.generate_random_noise(self.batch_shape),
                             dtype=torch.float32, device=self.device),
                requires_grad=True)
        else:
            if isinstance(init, np.ndarray):
                image = torch.tensor(init, device=device)
            elif isinstance(init, torch.Tensor):
                image = init.to(device)
            else:
                raise NotImplementedError(f"init must be np.ndarray or torch.Tensor, not {type(init)}")
            self.image = torch.nn.Parameter(image, requires_grad=True)
        self.init_optimizer()

    def get_image(self):
        return self.image

    def get_samples(self):
        return self.image

    def __getstate__(self):
        state = super().__getstate__()
        state.update({"image": self.image})
        return state

    def __setstate__(self, state):
        super().__setstate__(state)
        self.image = state["image"]


class MEI_variational(MEI_result):
    def __init__(self, distribution, img_shape, init=None, device='cpu', **MEIParams):
        assert "n_samples_per_batch" in MEIParams, "n_samples_per_batch must be specified"
        n_samples = MEIParams["n_samples_per_batch"]
        super(MEI_variational, self).__init__(img_shape, n_samples, device, **MEIParams)

        if distribution == "normal":
            dist = torch.distributions.Normal
        elif distribution == "laplace":
            dist = torch.distributions.Laplace
        elif distribution == "uniform":
            dist = torch.distributions.Uniform
        else:
            raise NotImplementedError(f"Distribution {distribution} not supported")

        assert "fixed_stddev" in MEIParams, "fixed_stddev must be specified"
        self.mean, self.std = self.generate_loc_scale(init, MEIParams["fixed_stddev"])
        self.register_parameter("mu", self.mean)
        if not MEIParams["fixed_stddev"]:
            self.register_parameter("sigma", self.std)
        self.distribution = dist(self.mean, self.std)

        self.init_optimizer()

    def get_image(self):
        return self.distribution.mean

    def get_samples(self):
        return self.distribution.rsample(torch.Size(self.n_samples))

    def generate_loc_scale(self, mean=None, fixed_stddev=False) -> (nn.Parameter, nn.Parameter):
        mean = self.generate_random_noise(self.img_shape) if mean is None else mean
        mean = torch.nn.Parameter(torch.tensor(mean,
                                               dtype=torch.float32,
                                               device=self.device),  requires_grad=True)
        
        stddev = fixed_stddev if fixed_stddev else 0.4
        std = torch.ones(self.img_shape, dtype=torch.float32, device=self.device) * stddev
        if not fixed_stddev:
            std = torch.nn.Parameter(std, requires_grad=True)
        return mean, std

    def __getstate__(self):
        state = super().__getstate__()
        state.update({"distribution": self.distribution})
        return state

    def __setstate__(self, state):
        super().__setstate__(state)
        self.distribution = state["distribution"]


class MEI_transformation(MEI_result):
    def __init__(self, transform, img_shape, device='cpu', **MEIParams):
        n_samples = MEIParams["n_samples"]
        del MEIParams["n_samples"]
        super(MEI_transformation, self).__init__(img_shape, n_samples, device, **MEIParams)

        self.transform = transform.to(self.device)
        self.batch_shape = (n_samples, *img_shape)

        self.init_optimizer()

    def get_image(self):
        return self.transform(self.n_samples, use_mean=True)

    def get_samples(self):
        return self.transform(self.n_samples)

    def __getstate__(self):
        state = super().__getstate__()
        state.update({"transform": self.transform})
        return state

    def __setstate__(self, state):
        super().__setstate__(state)
        self.transform = state["transform"]



from distutils.core import setup

try:
    from pypandoc import convert_text
except ImportError:
    convert_text = lambda text, *args, **kwargs: text


setup(
    name='meitorch',         # How you named your package folder (MyLib)
    packages=['meitorch', 'meitorch/objective', 'meitorch/tools'],   # Chose the same as "name"
    version='0.223',      # Start with a small number and increase it with every change you make
    license='MIT',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    description="Generate Most Exciting Input to explore and understand PyTorch model's behavior "
                "by identifying input samples that induce high activation from specific neurons in your model.",   # Give a short description about your library
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='László Freund',                   # Type in your name
    author_email='freundl0509@gmail.com',      # Type in your E-Mail
    url='https://github.com/lacykaltgr/mei',   # Provide either the link to your github or to your website
    download_url='https://github.com/lacykaltgr/mei/archive/refs/tags/v0.0.1.tar.gz',    # I explain this later on
    keywords=['generate', 'meitorch', 'most exciting', 'input optimization', 'interpretability'],   # Keywords that define your package best
    install_requires=[            # I get to this in a second
        'torch',
        'numpy',
        'scipy',
        'tqdm',
        'scikit-image',
        'torchvision',
        'matplotlib'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',      # Define that your audience are developers
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',   # Again, pick a license
        'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)


import numpy as np
import torch
from tqdm import tqdm
from scipy import optimize
from scipy import ndimage

from .objective.gabor import create_gabor_loader, create_gabor, default_gabor_ranges, default_gabor_limits
from .result import MEI_image


class LinearMEI:
    """
    Gabor filter creation, optimization, comparision
    Fewer features but a simpler interface then MEI
    """

    def __init__(self, operation, shape=(1, 28, 28), bias=0, scale=1, device='cpu'):
        self.bias = bias
        self.scale = scale
        self.device = device
        self.img_shape = shape
        self.operation = operation
        self.ranges: dict = default_gabor_ranges
        self.limits: list = default_gabor_limits
        self.set_gabor_ranges(height=[self.img_shape[-2]], width=[self.img_shape[-1]])

    def best_gabor(self, gabor_loader=None):
        """
        Find the most exciting gabor filter for cells in the neuron query

        :param gabor_loader: The dataset containing the gabor filters to be evaluated
        :return: MEI_image for best gabor filter for the queried neurons
        """
        if self.img_shape[0] != 1:
            raise ValueError("Only grayscale images are supported for this feature, try using optimal gabor")

        if gabor_loader is None:
            gabor_loader = create_gabor_loader(self.ranges)

        # Evaluate all gabors
        activations = []
        with torch.no_grad():
            print('Evaluating gabors')
            for i, gabor in tqdm(enumerate(gabor_loader)):
                norm = (gabor["image"] - self.bias) / self.scale
                img = torch.Tensor(norm).to(self.device)
                activations.append(self.operation(img).cpu().numpy())
        activations = np.array(activations)

        best_idx = np.argmax(activations)
        (_, _, best_phase, best_wavelength, best_orientation, best_sigma, best_dy,
         best_dx) = gabor_loader[best_idx]['params'].values()
        return MEI_image(
            n_images=1,
            shape=self.img_shape,
            init=gabor_loader[best_idx]['image'],
            bias=self.bias,
            scale=self.scale,
            device=self.device,
            activation=activations[best_idx],
            phase=best_phase,
            wavelength=best_wavelength,
            orientation=best_orientation,
            sigma=best_sigma,
            dy=best_dy,
            dx=best_dx)

    def optimal_gabor(self, target_mean=None, target_contrast=None):
        """
        Find parameters that produce an optimal gabor for this queried neurons

        :param target_mean: Target mean of optimal gabor
        :param target_contrast: Target contrast of optimal gabor
        :return: MEI_image for optimal gabor filters for the queried neurons
        """

        bounds = self.limits * self.img_shape[0] if self.img_shape[0] > 1 else self.limits

        def neg_model_activation(params):
            params = [np.clip(p, l, u) for p, (l, u) in zip(params, bounds)]
            gabor = [create_gabor(height=self.img_shape[-2], width=self.img_shape[-1],
                                  phase=params[6 * channel], wavelength=params[6 * channel + 1],
                                  orientation=params[6 * channel + 2],
                                  sigma=params[6 * channel + 3],
                                  dy=params[6 * channel + 4], dx=params[6 * channel + 5],
                                  target_mean=target_mean, target_contrast=target_contrast)
                     for channel in range(self.img_shape[0])]
            gabor = np.stack(gabor, axis=0)

            with torch.no_grad():
                norm = (gabor - self.bias) / self.scale
                img = torch.Tensor(norm).to(self.device)
                activation = self.operation(img).item()

            return -activation

        best_params = None
        best_activation = np.inf

        # Find the best parameters (simulated annealing -> local search)
        for seed in tqdm([1, 12, 123, 1234, 12345]):  # try 5 diff random seeds
            res = optimize.dual_annealing(neg_model_activation, bounds=bounds, no_local_search=True,
                                          maxiter=300, seed=seed)
            res = optimize.minimize(neg_model_activation, x0=res.x, method='Nelder-Mead')

            if res.fun < best_activation:
                best_activation = res.fun
                best_params = res.x
        if best_params is None:
            raise ValueError('No solution found')

        best_params = [np.clip(p, l, u) for p, (l, u) in zip(best_params, bounds)]

        # Create best gabor
        best_gabor = np.squeeze([create_gabor(height=self.img_shape[-2], width=self.img_shape[-1],
                                              phase=best_params[6 * i], wavelength=best_params[6 * i + 1],
                                              orientation=best_params[6 * i + 2], sigma=best_params[6 * i + 3],
                                              dy=best_params[6 * i + 4], dx=best_params[6 * i + 5])
                                 for i in range(self.img_shape[0])])
        best_activation = -neg_model_activation(best_params)

        return MEI_image(init=best_gabor,
                         shape=self.img_shape,
                         bias=self.bias,
                         scale=self.scale,
                         device=self.device,
                         activation=best_activation,
                         phase=[best_params[i] for i in range(self.img_shape[0])],
                         wavelength=[best_params[i + 1] for i in range(self.img_shape[0])],
                         orientation=[best_params[i + 2] for i in range(self.img_shape[0])],
                         sigma=[best_params[i + 3] for i in range(self.img_shape[0])],
                         dy=[best_params[i + 4] for i in range(self.img_shape[0])],
                         dx=[best_params[i + 5] for i in range(self.img_shape[0])])

    @staticmethod
    def white_noise_analysis(operation, shape, n_samples=1000, sigma=0.6, device="cpu"):
        if shape[0] == 1:
            shape = shape[1:]

        white_noise = np.random.normal(size=(n_samples, np.prod(shape)), loc=0.0, scale=1.)

        # apply ndimage.gaussian_filter with sigma=0.6
        for i in range(n_samples):
            white_noise[i, :] = ndimage.gaussian_filter(
                white_noise[i, :].reshape(shape), sigma=sigma).reshape(np.prod(shape))

        white_noise = torch.tensor(white_noise, dtype=torch.float32, device=device)
        values = []
        # loop over a batc h of 128 white_noise images
        with torch.no_grad():
            for i in range(0, n_samples, 128):
                batch = white_noise[i:i + 128, ...]
                batch = batch.reshape(-1, *shape)
                batch = torch.tensor(batch, dtype=torch.float32, device=device)
                values.append(operation(batch))
            values = torch.concatenate(values, dim=0)
        # multiply transpose of target block_values with white noise tensorially
        receptive_fields = torch.matmul(values.T, white_noise) / np.sqrt(n_samples)
        return receptive_fields

    def set_gabor_ranges(self, **ranges):
        """
        Ranges for the grid search for Gabor filter loader
        :param ranges: The ranges in key-value pairs
        """
        for key, value in ranges.items():
            self.ranges[key] = value

    def set_gabor_limits(self, **limits):
        """
        Bounds for Gabor filter optimization
        :param limits: The bounds in key-value pairs
        """

        def find_index(list_of_keys, element):
            for i, k in enumerate(list_of_keys):
                if k == element:
                    return i
            return -1

        for key, value in limits.items():
            index = find_index(self.ranges.keys(), key)
            if index == -1:
                raise ValueError("Invalid key")
            self.limits[index] = value



import kornia
import torch


def tv_norm(image):
    """
    # Calculate the TV norm
    # TV(x) = ||∇x||_1, where ∇x is the gradient of the image
    # You can use finite differences to calculate the gradient
    dx = torch.abs(image[:, :-1] - image[:, 1:])  # Differences along columns
    dy = torch.abs(image[:-1, :] - image[1:, :])  # Differences along rows

    # Sum the absolute differences to calculate the TV norm
    tv_norm = dx.sum() + dy.sum()
    """

    loss = kornia.losses.TotalVariation()
    return loss(image)


def l2_norm(image):
    return torch.norm(image, p=2)



from torch import optim
import torch
from ..tools.schedules import ConstantSchedule


def get_optimizer(
        params,
        optimizer_type: str = "adam",
        iter_n = None,
        **kwargs):
    """
    Get an optimizer based on the optimizer type and the learning rate

    :param params:
    :param optimizer_type: The optimizer type
    :param iter_n: The number of iterations
    :param kwargs: Additional arguments for the optimizer
    :return: The optimizer nn.Optimizer instance
    """
    if optimizer_type == "adam":
        return optim.Adam(params, **kwargs)
    elif optimizer_type == "sgd":
        return optim.SGD(params, **kwargs)
    elif optimizer_type == "rmsprop":
        return optim.RMSprop(params, **kwargs)
    elif optimizer_type == "mei":
        assert iter_n is not None, "iter_n must be specified"
        kwargs.update({"iter_n": iter_n})
        return MEIoptimizer(params, kwargs)
    elif optimizer_type == "meibatch":
        assert iter_n is not None, "iter_n must be specified"
        kwargs.update({"iter_n": iter_n})
        return MEIBatchoptimizer(params, kwargs)
    else:
        raise ValueError("Invalid optimizer type (must be 'adam', 'sgd' or 'rmsprop')")


class MEIoptimizer(optim.Optimizer):

    def __init__(self, params, defaults):
        assert "lr" in defaults, "lr must be specified"
        self.lr = defaults["lr"]
        super(MEIoptimizer, self).__init__(params, defaults)

        assert "iter_n" in defaults, "iter_n must be specified"
        self.iter_n = defaults["iter_n"]

        step_gain = defaults["step_gain"] if "step_gain" in defaults else 1
        if isinstance(step_gain, (int, float)):
            self.step_gain = ConstantSchedule(step_gain)(self.iter_n)
        elif callable(step_gain):
            self.step_gain = step_gain(self.iter_n)

        self.eps = defaults["eps"] if "eps" in defaults else 1e-8
        self.step_i = 0

    def step(self):
        step_gain = self.step_gain(self.step_i)
        step = None
        for param_group in self.param_groups:
            for param in param_group["params"]:
                if not param.requires_grad or param.grad is None:
                    continue
                grad = param.grad.data
                standard_deviation = 1  # hardcoded nem jó
                a = param_group["lr"] / (torch.abs(grad).mean() + self.eps)
                b = (step_gain / (2 * standard_deviation)) * grad.data
                step = a * b
                param.data -= step
        self.step_i += 1
        return step


class MEIBatchoptimizer(MEIoptimizer):
    def __init__(self, params, defaults):
        super(MEIBatchoptimizer, self).__init__(params, defaults)

    def step(self):
        step_gain = self.step_size(self.step_i)
        step = None
        for param_group in self.param_groups:
            for param in param_group["params"]:
                if not param.requires_grad or param.grad is None:
                    continue
                grad = param.grad.data
                standard_deviation = 1  # hardcoded nem jó
                a = param_group["lr"] / (torch.mean(torch.abs(grad.data), dim=0, keepdim=True) + self.eps)
                b = (step_gain / (2 * standard_deviation)) * grad.data
                step = a * b
                param.data -= step
        self.step_i += 1
        return step

# * both versions are equivalent for a single-image batch, for batches with more than
# one image the first one is better but it drawns out the gradients that are spatially
# wide; for instance a gradient of size 5 x 5 pixels all at amplitude 1 will produce a
# higher change in an image of the batch than a gradient of size 20 x 20 all at
# amplitude 1 in another. This is alright in most cases, but when generating diverse
# images with min linkage (i.e, all images receive gradient from the signal and two
# get the gradient from the diversity term) it drawns out the gradient generated from
# the diversity term (because it is usually bigger spatially than the signal gradient)
# and becomes hard to find very diverse images (i.e., increasing the diversity term
# has no effect because the diversity gradient gets rescaled down to smaller values
# than the signal gradient)
# In any way, gradient mean is only used as normalization here and using the mean is
# alright (also image generation works normally).
